clear all;
close all;
clc;

rand('twister',123);
randn('state',123);

%% PARAMETERS
nIterEM = 300; %number of iterations for EM algorithm
nIterCLL = 500; %number of iterations for CL optimization
nIterMaxMargin = 500; %number of iterations for max-margin optimization
nCompMax = 12; %number of components for GMM
updatePrior = 1; %learn class-prior as well
F1 = 7; %CL optimization parameter

%max-margin optimization parameters:
F2 = 15;
eta = 5;
sigma = 0.45; %can alternatively be set by cross-validation


%% Generate 2 Class Spiral - Training Samples
nTrainSampleClass1 = 10000;
nTrainSampleClass2 = 10000;
nTestSampleClass1 = 10000;
nTestSampleClass2 = 10000;

nTrainSample = nTrainSampleClass1 + nTrainSampleClass2;
nTestSample = nTestSampleClass1 + nTestSampleClass2;

Umdr=2;
a1=1;
sigN=1.1;

t1=2*pi*Umdr*rand(1,nTrainSampleClass1 + nTestSampleClass1);
t2=2*pi*Umdr*rand(1,nTrainSampleClass2 + nTestSampleClass2);
x1=[(a1.*t1).*cos(t1); (a1.*t1).*sin(t1)];
x2=[(-a1.*t2).*cos(t2); (-a1.*t2).*sin(t2)];

%Add zero mean Gaussian Noise
x1 = x1 + sigN.*randn(size(x1));
x2 = x2 + sigN.*randn(size(x2));


trainSample = [x1(:,1:nTrainSampleClass1) x2(:,1:nTrainSampleClass2)];
testSample = [x1(:,(nTrainSampleClass1+1):end) x2(:,(nTrainSampleClass2+1):end)];
trainLabel = [ones(1,nTrainSampleClass1) 2*ones(1,nTrainSampleClass2)];
testLabel = [ones(1,nTestSampleClass1) 2*ones(1,nTestSampleClass2)];


figure;
scatter(testSample(1,testLabel==1), testSample(2,testLabel==1));
hold on;
scatter(testSample(1,testLabel==2), testSample(2,testLabel==2),'r');
title('Test Samples');


%% Maximum likelihood class prior
classPrior_ML = [0.5 0.5];

%% Train GMM Classifier with EM Algorithm (we use the ML-solution for initialization of discriminative learning)
GMM_ML = cell(1,2);
options = statset('MaxIter',nIterEM);
for kClass = 1:2
    ind = (trainLabel == kClass);
    GMM_ML{kClass}=gmdistribution.fit(trainSample(:,ind).',nCompMax,'Start','randSample','CovType','diagonal','Replicates',10,'Options',options);
end

%% Convert GMM parameters to different format
GMM = cell(1,2);
for kClass = 1:2
    GMM{kClass} = cell(1,nCompMax);
    for kCompMax = 1:nCompMax
        GMM{kClass}{kCompMax}.prior = GMM_ML{kClass}.PComponents(kCompMax);
        GMM{kClass}{kCompMax}.mean = GMM_ML{kClass}.mu(kCompMax,:);
        GMM{kClass}{kCompMax}.cov = diag(GMM_ML{kClass}.Sigma(:,:,kCompMax));
    end
end

%% Train GMM Classifier using Conditional Likelihood)
[classPrior_CLL,GMM_CL,trainCR_CLL,testCR_CLL,trainCL_CLL] = maxCllGMM(trainSample.',trainLabel,nIterCLL,GMM,classPrior_ML,testSample.',testLabel,updatePrior,F1,1);


%% Train GMM Classifier using Max Margin
[classPrior_MM,GMM_MM,trainCR_MM,testCR_MM,trainCL_MM,trainMMApprox,trainMM,sampleMargin] = maxMarginGMM(trainSample.',trainLabel,nIterMaxMargin,GMM,classPrior_ML,sigma,testSample.',testLabel,eta,updatePrior,F2,1);


%% Plot performance results
figure;
subplot(3,1,1);
plot(trainCL_CLL);
grid on;
title('Conditional Likelihood cost function');

subplot(3,1,2);
plot(trainMMApprox);
grid on;
title('Max-Margin cost function (MM)');

subplot(3,1,3);
hold on;
plot(trainCR_CLL,'--');
plot(testCR_CLL);
plot(trainCR_MM,'r--');
plot(testCR_MM,'r');
grid on;
legend('CL-Train','CL-Test','MM-Train','MM-Test',0);
title('Classification Rate');

